"""CChecksum is a ~8x faster drop-in replacement for eth_utils.to_checksum_address,
with the most cpu-intensive part implemented in c.

It keeps the exact same API as the existing implementation, exceptions and all.
"""

from cchecksum._checksum import to_checksum_address, to_checksum_address_many
from cchecksum.monkey_patch import monkey_patch_eth_utils, monkey_patch_web3py

__all__ = [
    "to_checksum_address",
    "to_checksum_address_many",
    "monkey_patch_eth_utils",
    "monkey_patch_web3py",
]



import time
from binascii import unhexlify
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache

import eth_utils
import pytest

from cchecksum import to_checksum_address, to_checksum_address_many

benchmark_addresses = []
range_start = 100000000000000000000000000000000000000000


@lru_cache(maxsize=None)
def __get_prefix(addrlen: int) -> str:
    return f"0x{'0' * (40 - len(address))}"


for i in range(range_start, range_start + 500000):
    address = hex(i)[2:]
    address = f"{__get_prefix(len(address))}{address}"
    benchmark_addresses.append(address)


def test_checksum_str_lower() -> None:
    lower = "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2".lower()
    assert to_checksum_address(lower) == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"


def test_checksum_str_upper() -> None:
    lower = "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2".upper()
    assert to_checksum_address(lower) == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"


def test_checksum_str_mixed() -> None:
    lower = "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
    assert to_checksum_address(lower) == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"


def test_checksum_str_no_0x_prefix_lower() -> None:
    lower = "C02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2".lower()
    assert to_checksum_address(lower) == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"


def test_checksum_str_no_0x_prefix_upper() -> None:
    lower = "C02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2".upper()
    assert to_checksum_address(lower) == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"


def test_checksum_str_no_0x_prefix_mixed() -> None:
    lower = "C02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
    assert to_checksum_address(lower) == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"


def test_checksum_bytes() -> None:
    bytes = unhexlify("C02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2")
    assert to_checksum_address(bytes) == "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"


def test_checksum_many_str() -> None:
    addresses = benchmark_addresses[:10]
    expected = [to_checksum_address(addr) for addr in addresses]
    assert to_checksum_address_many(addresses) == expected


def test_checksum_many_bytes() -> None:
    addresses = benchmark_addresses[:10]
    addresses_bytes = [unhexlify(addr[2:]) for addr in addresses]
    expected = [to_checksum_address(addr) for addr in addresses_bytes]
    assert to_checksum_address_many(addresses_bytes) == expected


def test_checksum_many_packed_bytes() -> None:
    addresses = benchmark_addresses[:5]
    packed = b"".join(unhexlify(addr[2:]) for addr in addresses)
    expected = [to_checksum_address(addr) for addr in addresses]
    assert to_checksum_address_many(packed) == expected


def test_checksum_many_packed_memoryview() -> None:
    addresses = benchmark_addresses[:5]
    packed = b"".join(unhexlify(addr[2:]) for addr in addresses)
    expected = [to_checksum_address(addr) for addr in addresses]
    assert to_checksum_address_many(memoryview(packed)) == expected


def test_checksum_many_packed_length_error() -> None:
    with pytest.raises(ValueError, match="multiple of 20"):
        to_checksum_address_many(b"\x00" * 21)


def test_checksum_many_invalid_item_type() -> None:
    with pytest.raises(TypeError, match="Unsupported type"):
        to_checksum_address_many([1])


def test_type_error() -> None:
    input = object()
    try:
        eth_utils.to_checksum_address(input)
    except TypeError as e:
        with pytest.raises(TypeError, match=str(e)):
            to_checksum_address(input)
    else:
        raise RuntimeError("this should not happen")


def test_int_type_error() -> None:
    # this gives a ValueError in eth_utils but that's weird
    with pytest.raises(
        TypeError,
        match="Unsupported type: '<class 'int'>'. Must be one of: bool, str, bytes, bytearray or int.",
    ):
        to_checksum_address(0)


def test_none_type_error() -> None:
    # this gives a ValueError in eth_utils but that's due to an implementation detail of hexstr_to_str
    with pytest.raises(
        TypeError,
        match="Unsupported type: '<class 'NoneType'>'. Must be one of: bool, str, bytes, bytearray or int.",
    ):
        to_checksum_address(None)


def test_value_error() -> None:
    input = "i am a bad string"
    try:
        eth_utils.to_checksum_address(input)
    except ValueError as e:
        with pytest.raises(ValueError, match=str(e)):
            to_checksum_address(input)
    else:
        raise RuntimeError("this should not happen")


def test_benchmark() -> None:
    start = time.time()
    checksummed = list(map(to_checksum_address, benchmark_addresses))
    cython_duration = time.time() - start
    print(f"cython took {cython_duration}s")
    start = time.time()
    python = list(map(eth_utils.to_checksum_address, benchmark_addresses))
    python_duration = time.time() - start
    print(f"python took {python_duration}s")
    assert checksummed == python
    assert cython_duration < python_duration
    print(f"took {cython_duration/python_duration}% of the time")


def test_threadsafety() -> None:
    test_addresses = benchmark_addresses[:1000]
    executor = ThreadPoolExecutor(100)
    fast_way = executor.map(to_checksum_address, test_addresses)
    slow_way = executor.map(eth_utils.to_checksum_address, test_addresses)
    assert list(fast_way) == list(slow_way)



import pytest
from pytest_codspeed import BenchmarkFixture

from benchmarks.data import CONTAINER_CASES, CONTAINER_CASE_IDS, SINGLE_CASES, SINGLE_CASE_IDS
from cchecksum import to_checksum_address


@pytest.mark.benchmark(group="to_checksum_address_str")
@pytest.mark.parametrize("value", SINGLE_CASES, ids=SINGLE_CASE_IDS)
def test_to_checksum_address(benchmark: BenchmarkFixture, value: str) -> None:
    @benchmark
    def run_10k():
        for _ in range(10_000):
            to_checksum_address(value)


@pytest.mark.benchmark(group="to_checksum_address_bytes_container")
@pytest.mark.parametrize("values", CONTAINER_CASES, ids=CONTAINER_CASE_IDS)
def test_to_checksum_address_multi(benchmark: BenchmarkFixture, values: list[bytes]) -> None:
    @benchmark
    def run_container():
        for value in values:
            to_checksum_address(value)



import eth_utils
import eth_utils.address
import pytest

# pywin32 does not install properly on free-threaded python builds on Windows,
# so we skip tests invoking web3.py for Python3.13+ on Windows.
pytest.importorskip("web3")

import web3._utils as web3_utils
import web3.main as web3_main
import web3.middleware as web3_middleware

import cchecksum


def test_monkey_patch_eth_utils():
    cchecksum.monkey_patch_eth_utils()
    assert eth_utils.to_checksum_address is cchecksum.to_checksum_address
    assert eth_utils.address.to_checksum_address is cchecksum.to_checksum_address


def test_monkey_patch_web3py():
    cchecksum.monkey_patch_web3py()

    assert web3_main.to_checksum_address is cchecksum.to_checksum_address
    assert web3_utils.ens.to_checksum_address is cchecksum.to_checksum_address
    assert web3_utils.method_formatters.to_checksum_address is cchecksum.to_checksum_address
    assert web3_utils.normalizers.to_checksum_address is cchecksum.to_checksum_address
    assert web3_middleware.signing.to_checksum_address is cchecksum.to_checksum_address



from eth_typing import AnyAddress, ChecksumAddress, HexAddress
from typing import Iterable, List, Union

def to_checksum_address(value: Union[AnyAddress, str, bytes]) -> ChecksumAddress:
    """
    Convert an address to its EIP-55 checksum format.

    This function takes an address in any supported format and returns it in the
    checksummed format as defined by EIP-55. It uses a custom Cython implementation
    for the checksum conversion to optimize performance.

    Args:
        value: The address to be converted. It can be in any format supported by
            :func:`eth_utils.to_normalized_address`.

    Raises:
        ValueError: If the input address is not in a recognized format.
        TypeError: If the input is not a string, bytes, or any address type.

    Examples:
        >>> to_checksum_address("0xb47e3cd837ddf8e4c57f05d70ab865de6e193bbb")
        '0xb47e3cd837dDF8e4c57F05d70Ab865de6e193BBB'

        >>> to_checksum_address(b'\xb4~<\xd87\xdd\xf8\xe4\xc5\x7f\x05\xd7\n\xb8e\xden\x19;\xbb')
        '0xb47e3cd837dDF8e4c57F05d70Ab865de6e193BBB'

    See Also:
        - :func:`eth_utils.to_checksum_address` for the standard implementation.
        - :func:`to_normalized_address` for converting to a normalized address before checksumming.
    """

def to_checksum_address_many(
    values: Union[
        bytes,
        bytearray,
        memoryview,
        Iterable[Union[AnyAddress, str, bytes, bytearray]],
    ],
) -> List[ChecksumAddress]:
    """
    Convert multiple addresses to EIP-55 checksum format.

    Accepts a sequence of address-like inputs (str/bytes/bytearray) or a packed
    bytes-like object containing concatenated 20-byte addresses.
    """

def to_normalized_address_no_0x(value: Union[AnyAddress, str, bytes]) -> bytes:
    """
    Converts an address to its normalized hexadecimal representation without the '0x' prefix.

    This function ensures that the address is in a consistent lowercase hexadecimal
    format, which is useful for further processing or validation.

    Args:
        value: The address to be normalized.

    Raises:
        ValueError: If the input address is not in a recognized format.
        TypeError: If the input is not a string, bytes, or any address type.

    Examples:
        >>> to_normalized_address("0xB47E3CD837DDF8E4C57F05D70AB865DE6E193BBB")
        '0xb47e3cd837ddf8e4c57f05d70ab865de6e193bbb'

        >>> to_normalized_address(b'\xb4~<\xd87\xdd\xf8\xe4\xc5\x7f\x05\xd7\n\xb8e\xden\x19;\xbb')
        '0xb47e3cd837ddf8e4c57f05d70ab865de6e193bbb'

    See Also:
        - :func:`eth_utils.to_normalized_address` for the standard implementation.
    """

def cchecksum(norm_address_no_0x: bytes, address_hash_hex_no_0x: bytes) -> ChecksumAddress: ...



from pathlib import Path
from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

classifiers=[
    "Intended Audience :: Developers",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "Programming Language :: Python :: 3.14",
    "Programming Language :: Python :: Implementation :: CPython",
    "Operating System :: OS Independent",
    "Topic :: Software Development :: Libraries",
]

with open("requirements.txt", "r") as f:
    requirements = list(map(str.strip, f.read().split("\n")))[:-1]

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

sources = ["cchecksum/_checksum.pyx", "cchecksum/keccak.c"]
extension = Extension("cchecksum._checksum", sources=sources, include_dirs=["cchecksum"])


class BuildExt(build_ext):
    def build_extensions(self):
        compiler_type = self.compiler.compiler_type
        opt_flag = "/O2" if compiler_type == "msvc" else "-O3"
        for ext in self.extensions:
            extra = list(ext.extra_compile_args or [])
            if opt_flag not in extra:
                extra.append(opt_flag)
            ext.extra_compile_args = extra
        super().build_extensions()


setup(
    name="cchecksum",
    packages=find_packages(),
    version="0.4.1",
    description="An ~18x faster drop-in replacement for eth_utils.to_checksum_address. Raises the exact same Exceptions. Implemented in C.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="BobTheBuidler",
    author_email="bobthebuidlerdefi@gmail.com",
    url="https://github.com/BobTheBuidler/cchecksum",
    license="MIT",
    install_requires=requirements,
    python_requires=">=3.9,<4",
    package_data={"cchecksum": ["py.typed", "*.pxd", "**/*.pxd"]},
    include_package_data=True,
    classifiers=classifiers,
    ext_modules=cythonize([extension], compiler_directives={"language_level": 3, "embedsignature": True, "linetrace": False, "cdivision": True, "initializedcheck": False, "nonecheck": False, "boundscheck": False, "wraparound": False}),
    zip_safe=False,
    cmdclass={"build_ext": BuildExt},
)



from cchecksum._checksum import to_checksum_address


def monkey_patch_eth_utils() -> None:
    """Monkey patch eth_utils to use cchecksum's implementation internally."""
    import eth_utils
    import eth_utils.address

    eth_utils.to_checksum_address = to_checksum_address
    eth_utils.address.to_checksum_address = to_checksum_address


def monkey_patch_web3py() -> None:
    """Monkey patch web3.py to use cchecksum's implementation internally."""
    import web3._utils as web3_utils
    import web3.main as web3_main
    import web3.middleware as web3_middleware

    web3_main.to_checksum_address = to_checksum_address
    web3_utils.ens.to_checksum_address = to_checksum_address
    web3_utils.method_formatters.to_checksum_address = to_checksum_address
    web3_utils.normalizers.to_checksum_address = to_checksum_address
    web3_middleware.signing.to_checksum_address = to_checksum_address

    try:
        import web3.utils.address as web3_address

        web3_address.to_checksum_address = to_checksum_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass

    try:
        import ens.ens

        ens.ens.to_checksum_address = to_checksum_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass

    try:
        import ens.async_ens

        ens.async_ens.to_checksum_address = to_checksum_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass



from binascii import unhexlify
from typing import Final

BASE_ADDRESSES: Final = [
    "0x52908400098527886e0f7030069857d2e4169ee7",
    "0xde709f2102306220921060314715629080e2fb77",
    "0x27b1fdb04752bbc536007a920d24acb045561c26",
    "0x5aeda56215b167893e80b4fe645ba6d5bab767de",
    "0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2",
]

MIXED_ADDRESSES: Final = [
    "0x52908400098527886E0F7030069857D2E4169EE7",
    "0x8617E340B3D01FA5F11F306F4090FD50E238070D",
    "0x5aAeb6053F3E94C9b9A09f33669435E7Ef1BeAed",
    "0xfb6916095ca1df60bb79ce92ce3ea74c37c5d359",
    "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
]

STR_CASES: Final = []
STR_CASE_IDS: Final = []
for index, address in enumerate(BASE_ADDRESSES):
    STR_CASES.append(address)
    STR_CASE_IDS.append(f"lower-0x-{index}")

    STR_CASES.append(address.upper())
    STR_CASE_IDS.append(f"upper-0x-{index}")

    STR_CASES.append(address[2:])
    STR_CASE_IDS.append(f"lower-no-0x-{index}")

    STR_CASES.append(address[2:].upper())
    STR_CASE_IDS.append(f"upper-no-0x-{index}")

for index, address in enumerate(MIXED_ADDRESSES):
    STR_CASES.append(address)
    STR_CASE_IDS.append(f"mixed-{index}")

BYTES_CASES: Final = [unhexlify(address[2:]) for address in BASE_ADDRESSES]
BYTES_CASE_IDS: Final = [f"bytes-{index}" for index in range(len(BASE_ADDRESSES))]

BYTEARRAY_CASES: Final = [bytearray(address) for address in BYTES_CASES]
BYTEARRAY_CASE_IDS: Final = [f"bytearray-{index}" for index in range(len(BYTES_CASES))]

SINGLE_CASES: Final = STR_CASES + BYTES_CASES + BYTEARRAY_CASES
SINGLE_CASE_IDS: Final = STR_CASE_IDS + BYTES_CASE_IDS + BYTEARRAY_CASE_IDS


CONTAINER_SIZES: Final = 100, 1_000, 10_000


def _build_container(cases: list, size: int) -> list:
    return [cases[index % len(cases)] for index in range(size)]


STR_CONTAINER_CASES: Final = [_build_container(STR_CASES, size) for size in CONTAINER_SIZES]
STR_CONTAINER_CASE_IDS: Final = [f"str-container-{size}" for size in CONTAINER_SIZES]

BYTES_CONTAINER_CASES: Final = [_build_container(BYTES_CASES, size) for size in CONTAINER_SIZES]
BYTES_CONTAINER_CASE_IDS: Final = [f"bytes-container-{size}" for size in CONTAINER_SIZES]

BYTEARRAY_CONTAINER_CASES: Final = [
    _build_container(BYTEARRAY_CASES, size) for size in CONTAINER_SIZES
]
BYTEARRAY_CONTAINER_CASE_IDS: Final = [f"bytearray-container-{size}" for size in CONTAINER_SIZES]

CONTAINER_CASES: Final = STR_CONTAINER_CASES + BYTES_CONTAINER_CASES + BYTEARRAY_CONTAINER_CASES
CONTAINER_CASE_IDS: Final = (
    STR_CONTAINER_CASE_IDS + BYTES_CONTAINER_CASE_IDS + BYTEARRAY_CONTAINER_CASE_IDS
)



from cpython.bytes cimport PyBytes_GET_SIZE
from cpython.sequence cimport PySequence_Fast, PySequence_Fast_GET_ITEM, PySequence_Fast_GET_SIZE
from cpython.unicode cimport PyUnicode_AsEncodedString, PyUnicode_DecodeASCII
from cython.parallel cimport prange
from libc.stddef cimport size_t
from libc.string cimport memcpy

from eth_typing import AnyAddress, ChecksumAddress


cdef const unsigned char* hexdigits = b"0123456789abcdef"


# this was ripped out of eth_utils and optimized a little bit

cdef extern from "keccak.h":
    void keccak_256(const unsigned char* data, size_t len, unsigned char* out) nogil


cpdef unicode to_checksum_address(value: Union[AnyAddress, str, bytes]):
    """
    Convert an address to its EIP-55 checksum format.

    This function takes an address in any supported format and returns it in the
    checksummed format as defined by EIP-55. It uses a custom Cython implementation
    for the checksum conversion to optimize performance.

    Args:
        value: The address to be converted. It can be in any format supported by
            :func:`eth_utils.to_normalized_address`.

    Raises:
        ValueError: If the input address is not in a recognized format.
        TypeError: If the input is not a string, bytes, or any address type.

    Examples:
        >>> to_checksum_address("0xb47e3cd837ddf8e4c57f05d70ab865de6e193bbb")
        '0xb47e3cd837dDF8e4c57F05d70Ab865de6e193BBB'

        >>> to_checksum_address(b'\xb4~<\xd87\xdd\xf8\xe4\xc5\x7f\x05\xd7\n\xb8e\xden\x19;\xbb')
        '0xb47e3cd837dDF8e4c57F05d70Ab865de6e193BBB'

    See Also:
        - :func:`eth_utils.to_checksum_address` for the standard implementation.
        - :func:`to_normalized_address` for converting to a normalized address before checksumming.
    """
    cdef bytes hex_address_bytes
    cdef const unsigned char* hex_address_bytestr
    cdef unsigned char c
    cdef unsigned char hash_out[32]

    # Create a buffer for our result
    # 2 for "0x" prefix and 40 for the address itself
    cdef char result_buffer[42]
    result_buffer[0] = 48  # '0'
    result_buffer[1] = 120  # 'x'
    
    if isinstance(value, str):
        hex_address_bytes = lowercase_ascii_and_validate(PyUnicode_AsEncodedString(value, b"ascii", NULL))            
        hex_address_bytestr = hex_address_bytes

    elif isinstance(value, (bytes, bytearray)):
        hex_address_bytes = hexlify(value)
        hex_address_bytestr = hex_address_bytes
        num_bytes = PyBytes_GET_SIZE(hex_address_bytes)

        with nogil:
            for i in range(num_bytes):
                c = hex_address_bytestr[i]
                
                if not is_hex_lower(c):
                    raise ValueError(
                        f"Unknown format {repr(value)}, attempted to normalize to '0x{hex_address_bytes.decode()}'"
                    )
        
    else:
        raise TypeError(
            f"Unsupported type: '{repr(type(value))}'. Must be one of: bool, str, bytes, bytearray or int."
        )

    if PyBytes_GET_SIZE(hex_address_bytes) != 40:
        raise ValueError(
            f"Unknown format {repr(value)}, attempted to normalize to '0x{hex_address_bytes.decode()}'"
        )
    
    with nogil:
        keccak_256(hex_address_bytestr, 40, hash_out)
        populate_result_buffer(result_buffer, hex_address_bytestr, hash_out)
        
    # It is faster to decode a buffer with a known size ie buffer[:42]
    return result_buffer[:42].decode('ascii')


cpdef list to_checksum_address_many(object values):
    """
    Convert multiple addresses to EIP-55 checksum format.

    Accepts a sequence of address-like inputs (str/bytes/bytearray) or a packed
    bytes-like object containing concatenated 20-byte addresses.
    """
    cdef Py_ssize_t i, n, packed_len
    cdef bytes hex_address_bytes
    cdef const unsigned char* hex_address_bytestr
    cdef object item
    cdef object seq
    cdef const unsigned char[:] packed_view
    cdef const unsigned char* packed_ptr
    cdef unsigned char* norm_ptr
    cdef unsigned char* hash_ptr
    cdef char* result_ptr
    cdef bytearray norm_hex
    cdef bytearray hashes
    cdef bytearray results
    cdef list output

    if isinstance(values, (bytes, bytearray, memoryview)):
        packed_view = values

        if packed_view.ndim != 1 or packed_view.itemsize != 1 or packed_view.strides[0] != 1:
            raise TypeError("Packed addresses must be a contiguous 1-D view of bytes.")

        packed_len = packed_view.shape[0]
        if packed_len % 20 != 0:
            raise ValueError("Packed address bytes length must be a multiple of 20.")

        n = packed_len // 20
        if n == 0:
            return []

        norm_hex = bytearray(n * 40)
        hashes = bytearray(n * 32)
        results = bytearray(n * 42)
        norm_ptr = norm_hex
        hash_ptr = hashes
        result_ptr = results
        packed_ptr = &packed_view[0]

        with nogil:
            for i in range(n):
                hexlify_c_string_to_buffer(packed_ptr + (i * 20), norm_ptr + (i * 40), 20)

            for i in prange(n, schedule="static"):
                keccak_256(norm_ptr + (i * 40), 40, hash_ptr + (i * 32))
                checksum_address_to_buffer(
                    result_ptr + (i * 42),
                    norm_ptr + (i * 40),
                    hash_ptr + (i * 32),
                )

        output = [None] * n
        for i in range(n):
            output[i] = PyUnicode_DecodeASCII(result_ptr + (i * 42), 42, NULL)
        return output

    if isinstance(values, str):
        raise TypeError("to_checksum_address_many expects a sequence of addresses, not a str.")

    seq = PySequence_Fast(values, "to_checksum_address_many expects a sequence of addresses.")
    n = PySequence_Fast_GET_SIZE(seq)
    if n == 0:
        return []

    norm_hex = bytearray(n * 40)
    hashes = bytearray(n * 32)
    results = bytearray(n * 42)
    norm_ptr = norm_hex
    hash_ptr = hashes
    result_ptr = results

    for i in range(n):
        item = <object>PySequence_Fast_GET_ITEM(seq, i)

        if isinstance(item, str):
            hex_address_bytes = lowercase_ascii_and_validate(PyUnicode_AsEncodedString(item, b"ascii", NULL))
            hex_address_bytestr = hex_address_bytes
        elif isinstance(item, (bytes, bytearray)):
            hex_address_bytes = hexlify(item)
            hex_address_bytestr = hex_address_bytes
        else:
            raise TypeError(
                f"Unsupported type: '{repr(type(item))}'. Must be one of: bool, str, bytes, bytearray or int."
            )

        if PyBytes_GET_SIZE(hex_address_bytes) != 40:
            raise ValueError(
                f"Unknown format {repr(item)}, attempted to normalize to '0x{hex_address_bytes.decode()}'"
            )

        memcpy(norm_ptr + (i * 40), hex_address_bytestr, 40)

    with nogil:
        for i in prange(n, schedule="static"):
            keccak_256(norm_ptr + (i * 40), 40, hash_ptr + (i * 32))
            checksum_address_to_buffer(
                result_ptr + (i * 42),
                norm_ptr + (i * 40),
                hash_ptr + (i * 32),
            )

    output = [None] * n
    for i in range(n):
        output[i] = PyUnicode_DecodeASCII(result_ptr + (i * 42), 42, NULL)
    return output


cpdef bytes hexlify(const unsigned char[:] src_buffer):
    return bytes(hexlify_unsafe(src_buffer, len(src_buffer)))


cdef const unsigned char[:] hexlify_unsafe(const unsigned char[:] src_buffer, Py_ssize_t num_bytes) noexcept:
    """Make sure your `num_bytes` is correct or ting go boom"""
    cdef unsigned char[:] result_buffer = bytearray(num_bytes * 2)  # contiguous and writeable
    with nogil:
        hexlify_memview_to_buffer(src_buffer, result_buffer, num_bytes)
    return result_buffer


cdef inline void hexlify_memview_to_buffer(
    const unsigned char[:] src_buffer, 
    unsigned char[:] result_buffer, 
    Py_ssize_t num_bytes,
) noexcept nogil:
    cdef Py_ssize_t i
    cdef unsigned char c
    for i in range(num_bytes):
        c = src_buffer[i]
        result_buffer[2*i] = hexdigits[c >> 4]
        result_buffer[2*i+1] = hexdigits[c & 0x0F]


cdef inline void hexlify_c_string_to_buffer(
    const unsigned char* src_buffer, 
    unsigned char* result_buffer, 
    Py_ssize_t num_bytes,
) noexcept nogil:
    cdef Py_ssize_t i
    cdef unsigned char c
    for i in range(num_bytes):
        c = src_buffer[i]
        result_buffer[2*i] = hexdigits[c >> 4]
        result_buffer[2*i+1] = hexdigits[c & 0x0F]


cdef inline void checksum_address_to_buffer(
    char* buffer,
    const unsigned char* norm_address_no_0x,
    const unsigned char* address_hash_bytes,
) noexcept nogil:
    cdef Py_ssize_t i
    cdef unsigned char hash_byte
    cdef unsigned char hash_nibble
    cdef unsigned char c

    buffer[0] = 48  # '0'
    buffer[1] = 120  # 'x'

    for i in range(40):
        c = norm_address_no_0x[i]
        hash_byte = address_hash_bytes[i >> 1]
        if i & 1:
            hash_nibble = hash_byte & 0x0F
        else:
            hash_nibble = hash_byte >> 4
        if hash_nibble < 8:
            buffer[i + 2] = c
        else:
            buffer[i + 2] = get_char(c)


cdef void populate_result_buffer(
    char[42] buffer,
    const unsigned char* norm_address_no_0x, 
    const unsigned char* address_hash,
) noexcept nogil:
    """
    Computes the checksummed version of an Ethereum address.

    This function takes a normalized Ethereum address (without the '0x' prefix) and its corresponding
    raw keccak hash bytes and returns the checksummed address as per the Ethereum Improvement
    Proposal 55 (EIP-55).

    Args:
        norm_address_no_0x: The normalized Ethereum address without the '0x' prefix.
        address_hash: The raw keccak hash bytes for the normalized address.

    Returns:
        The checksummed Ethereum address with the '0x' prefix.

    See Also:
        - :func:`eth_utils.to_checksum_address`: A utility function for converting addresses to their checksummed form.
    """
    cdef Py_ssize_t i
    cdef Py_ssize_t address_index
    cdef Py_ssize_t buffer_index
    cdef unsigned char hash_byte
    cdef unsigned char high_nibble
    cdef unsigned char low_nibble

    for i in range(20):
        hash_byte = address_hash[i]
        high_nibble = hash_byte >> 4
        low_nibble = hash_byte & 0x0F
        address_index = i * 2
        buffer_index = address_index + 2

        if high_nibble < 8:
            buffer[buffer_index] = norm_address_no_0x[address_index]
        else:
            buffer[buffer_index] = get_char(norm_address_no_0x[address_index])

        if low_nibble < 8:
            buffer[buffer_index + 1] = norm_address_no_0x[address_index + 1]
        else:
            buffer[buffer_index + 1] = get_char(norm_address_no_0x[address_index + 1])


cdef inline unsigned char get_char(unsigned char c) noexcept nogil:
    """This checks if `address_char` falls in the ASCII range for lowercase hexadecimal
    characters ('a' to 'f'), which correspond to ASCII values 97 to 102. If it does,
    the character is capitalized.
    """
    if c == 97:     # a
        return 65   # A
    elif c == 98:   # b
        return 66   # B
    elif c == 99:   # c
        return 67   # C
    elif c == 100:  # d
        return 68   # D
    elif c == 101:  # e
        return 69   # E
    elif c == 102:  # f
        return 70   # F
    else:
        return c


cdef inline bint is_hex_lower(unsigned char c) noexcept nogil:
    return (48 <= c <= 57) or (97 <= c <= 102)


cdef unsigned char* lowercase_ascii_and_validate(bytes src):
    cdef Py_ssize_t src_len, range_start, i
    cdef unsigned char* c_string
    cdef unsigned char c
    
    src_len = PyBytes_GET_SIZE(src)
    c_string = src

    with nogil:
        # if c_string[0] == b"0" and c_string[1] in (b"X", b"x")
        if c_string[0] == 48 and c_string[1] in (88, 120):
            range_start = 2
        else:
            range_start = 0
    
        for i in range(range_start, src_len):
            c = c_string[i]

            if 65 <= c <= 90:
                c += 32
                c_string[i] = c

            if c == 48:  # 0
                pass
            elif c == 49:  # 1
                pass
            elif c == 50:  # 2
                pass
            elif c == 51:  # 3
                pass
            elif c == 52:  # 4
                pass
            elif c == 53:  # 5
                pass
            elif c == 54:  # 6
                pass
            elif c == 55:  # 7
                pass
            elif c == 56:  # 8
                pass
            elif c == 57:  # 9
                pass
            elif c == 97:  # a
                pass
            elif c == 98:  # b
                pass
            elif c == 99:  # c
                pass
            elif c == 100:  # d
                pass
            elif c == 101:  # e
                pass
            elif c == 102:  # f
                pass
            else:
                with gil:
                    raise ValueError("when sending a str, it must be a hex string. " f"Got: {repr(src.decode('ascii'))}")
    
    return c_string[range_start:]


del AnyAddress, ChecksumAddress
